# -*- coding:utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

# ----------------------------------------------------------
# Author: Stephen Leger (s-leger)
#
# ----------------------------------------------------------
# noinspection PyUnresolvedReferences
import bpy
major, minor, rev = bpy.app.version
# noinspection PyUnresolvedReferences
from bpy.props import (
    BoolProperty,
    StringProperty,
    EnumProperty,
    IntVectorProperty,
    IntProperty,
    FloatProperty
)
from mathutils import Vector, Matrix
from mathutils.geometry import (
    intersect_line_plane
    )
from math import floor
from bpy_extras import view3d_utils
from .archipack_iconmanager import icons
from .archipack_abstraction import ArchipackObjectsManager, ensure_select_and_restore
from random import seed, uniform
from .archipack_prefs import get_prefs
objman = ArchipackObjectsManager()  
    
    
import logging
import time
logger = logging.getLogger(__package__)


Z_AXIS = Vector((0, 0, 1))


class stop_auto_manipulate():
    """Context Manager

    Usage:

    with stop_auto_manipulate(context):
        bpy.ops.archipack ...
    :param context: blender context
    """

    def __init__(self, context):
        if hasattr(context.window_manager, "archipack"):
            wm = context.window_manager.archipack
            self._ctx = context
            self._state = wm.auto_manipulate
            wm.auto_manipulate = False

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self._ctx.window_manager, "archipack"):
         self._ctx.window_manager.archipack.auto_manipulate = self._state

        # false let raise exception if any
        return None


def get_enum(self, context):
    return icons.enum(context, self.__class__.__name__, "json")


def preset_operator(self, context):
    import os
    base_preset, ext = os.path.splitext(self.preset)
    bpy.ops.archipack.preset_loader(filepath="{}.json".format(base_preset))


class ArchipackPanel:
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Archipack'
    bl_translation_context = "archipack"

    def draw_common(self, context, layout, draw_animation=True):
        if hasattr(context.window_manager, "archipack"):
            wm = context.window_manager.archipack
            self.draw_translate(context, layout)

            row = layout.row(align=True)
            self.draw_prop(context, layout, row, wm, "auto_manipulate", icon='AUTO', emboss=True)
            if not wm.auto_manipulate:
                self.draw_op(context, layout, row, "archipack.manipulate", icon='VIEW_PAN')
            self.draw_op(context, layout, layout, "archipack.delete", icon="TRASH")

        # support for animation -> disabled as there is no context available while rendering
        prefs = get_prefs(context)
        if prefs.experimental_features and draw_animation:
            o = context.active_object
            if o is not None and hasattr(context.scene, "archipack_animation"):
                row = layout.row(align=True)
                self.draw_prop(context, layout, row, o, "archipack_animation", icon="ACTION", emboss=True)
                icon = "HIDE_ON"
                if context.scene.archipack_animation:
                    icon = "HIDE_OFF"
                self.draw_prop(context, layout, row, context.scene, "archipack_animation", icon=icon)


class ArchipackGenericOperator(ArchipackObjectsManager):
    """
     Generic operator working on any archipack object
     Provide generic datablock accessor
     Polling for selected and active archipack objects
    """
    bl_options = {'INTERNAL', 'UNDO'}
    translation_context = "archipack"

    @classmethod
    def poll(cls, context):
        o = context.active_object
        return o and ArchipackObjectsManager.is_selected(cls, o) and cls.filter(o)

    @classmethod
    def filter(cls, o):
        return cls.archipack_filter(o)

    def datablock(self, o):
        """
         Return archipack datablock from object
        """
        return self.archipack_datablock(o)


class ArchipackObject(ArchipackObjectsManager):
    """
        Shared property of archipack's objects PropertyGroup
        provide basic support for copy to selected
        and datablock access / filtering by object
    """
    bl_translation_context = "archipack"

    def __init__(self):
        self.previously_selected = []
        self.previously_active = None
        self.auto_manipulate_state = True

    @property
    def random_color(self):
        return Vector((uniform(0.0, 1.0), uniform(0.0, 1.0), uniform(0.0, 1.0), 1))

    def iskindof(self, o, typ):
        """
            return true if object contains databloc of typ name
        """
        return o.data is not None and typ in o.data

    @property
    def category(self):
        return self.__class__.__name__[10::]

    @classmethod
    def filter(cls, o):
        """
            Filter object with this class in data
            return
            True when object contains this datablock
            False otherwhise
            usage:
            class_name.filter(object) from outside world
            self.__class__.filter(object) from instance
        """
        res = False
        try:
            res = cls.__name__ in o.data and getattr(o.data, cls.__name__)[0]
        except:
            pass
        if not res:
            try:
                res = cls.__name__ in o and getattr(o, cls.__name__)[0]
            except:
                pass
        return res

    @classmethod
    def poll(cls, o):
        """
            Filter object with this class in data
            return
            True when object contains this datablock
            False otherwhise
            usage:
            class_name.filter(object) from outside world
            self.__class__.filter(object) from instance
        """
        res = False
        try:
            res = ArchipackObjectsManager.is_selected(cls, o) and cls.filter(o)
        except:
            pass
        return res

    @classmethod
    def datablock(cls, o):
        """
            Retrieve datablock from base object
            return
                datablock when found
                None when not found
            usage:
                class_name.datablock(object) from outside world
                self.__class__.datablock(object) from instance
        """
        d = None
        try:
            d = getattr(o.data, cls.__name__)[0]
        except:
            pass

        if d is None:
            try:
                d = getattr(o, cls.__name__)[0]
            except:
                pass

        return d

    def on_delete(self, context, obj):
        """ Default on_delete, delete selected objects of same kind
        :param context:
        :param obj:
        :return:
        """
        for o in context.selected_objects:
            if self.filter(o) and o.name != obj.name:
                self.delete_object(context, o)

    def disable_auto_manipulate(self, context):
        if hasattr(context.window_manager, "archipack"):
            wm = context.window_manager.archipack
            # 2nd call on instances ??
            self.auto_manipulate_state = wm.auto_manipulate
            wm.auto_manipulate = False
            try:
                # fails at render time
                if wm.update_cursor:
                    context.window.cursor_set("WAIT")
            except:
                pass

    def restore_auto_manipulate(self, context):
        if hasattr(context.window_manager, "archipack"):
            wm = context.window_manager.archipack
            if self.auto_manipulate_state:
                wm.auto_manipulate = self.auto_manipulate_state
            try:
                if wm.update_cursor:
                    context.window.cursor_set("DEFAULT")
            except:
                pass

    def find_in_selection(self, context, auto_update=True):
        """
            find witch selected object this datablock instance belongs to
            store context to be able to restore after oops
            provide support for "copy to selected"
            XXX provide support for "context.object" override and preserve active one
            return
            object or None when instance not found in selected objects
            
            context.active_object is None when object's layer is hidden
            context.object does work in any case

            In render context there is no selection available
        """
        if auto_update is False:
            return None
        o = self.get_active_object(context)
        # set by context.view_layer.objects.active
        if hasattr(context, "object"):
            # set by user selection
            active = context.active_object
            selected = context.selected_objects[:]
        else:
            active = o
            selected = [o]


        self.previously_selected = []
        self.previously_active = None

        if self.__class__.datablock(o) == self:
            self.previously_selected = selected
            self.previously_active = active
            self.disable_auto_manipulate(context)
            return o

        for o in selected:
            if self.__class__.datablock(o) == self:
                self.previously_selected = selected
                self.previously_active = active
                self.disable_auto_manipulate(context)
                return o

        return None

    def restore_context(self, context):
        """
         restore context
        """
        # context.view_layer.update()
        bpy.ops.object.select_all(action="DESELECT")

        try:
            for o in self.previously_selected:
                self.select_object(context, o)
        except:
            pass

        if self.previously_active is not None:
            logger.debug("restore_context: %s", self.previously_active.name)

        self.select_object(context, self.previously_active, True)
        self.restore_auto_manipulate(context)
        self.previously_selected = None
        self.previously_active = None

    def id_mat(self, index, source=None):
        """ Limit material index to number of materials
        :param index:
        :param source: source obj with .idmat array, allow override on parent from childs
        :return:
        """
        if source is None:
            _source = self
        else:
            _source = source
        _idx = _source.idmat[index]
        if _idx < len(self.id_data.materials):
            return _idx
        return 0

    def shade_smooth(self, context, o, smooth_angle):
        m = o.data
        if not m.use_auto_smooth or m.auto_smooth_angle != smooth_angle:
            m.use_auto_smooth = True
            m.auto_smooth_angle = smooth_angle
        t = time.time()
        with ensure_select_and_restore(context, o, [o]):
            bpy.ops.object.shade_smooth()
        # logger.debug("bpy.ops.object.shade_smooth %.4f" % (time.time() - t))

    def find_modifier_by_type(self, o, mod_type):
        for mod in o.modifiers:
            if mod.type == mod_type:
                return mod
        return None

    def bevel_modifier(self, o, width=0.0005):
        mod = self.find_modifier_by_type(o, "BEVEL")
        if mod is None:
            mod = o.modifiers.new("Bevel", "BEVEL")
        mod.limit_method = 'ANGLE'
        mod.offset_type = 'WIDTH'
        mod.width = width
        mod.harden_normals = True
        mod.segments = 4
        mod.use_clamp_overlap = False


class ArchipackCreateTool(ArchipackObjectsManager):
    """
        Shared property of archipack's create tool Operator
    """
    bl_category = 'Archipack'
    bl_options = {'INTERNAL', 'UNDO'}

    filepath: StringProperty(
        options={'SKIP_SAVE'},
        name="Preset",
        description="Full filename of python preset to load at create time",
        default=""
    )

    act = None

    @property
    def archipack_category(self):
        """
            return target object name from ARCHIPACK_OT_object_name
        """
        return self.bl_idname[13:]

    def load_preset(self, d, auto_update=True):
        """
            Load python preset
            d: archipack object datablock
            preset: full filename.py with path
        """

        d.auto_update = False

        if self.filepath != "":
            try:
                # print("Load preset")
                bpy.ops.archipack.preset_loader(filepath=self.filepath)
                # print("Load preset success")
            except:
                logger.debug("Archipack unable to load preset file : %s" % (self.filepath))
                pass

        d.auto_update = auto_update

    def add_material(self, context, o, material='', category=None):
        # skip if preset allready add material
        if "archipack_material" in o:
            return
        # enable viewport transparency
        o.show_transparent = True
        try:
            if category is None:
                category = self.archipack_category

            if bpy.ops.archipack.material.poll():
                bpy.ops.archipack.material(category=category, material=material)

        except Exception as ex:
            import traceback
            traceback.print_exc()
            logger.debug("Archipack %s materials not found" % (self.archipack_category))
            pass

    def invoke(self, context, event):
        o = context.active_object
        if o is None:
            # fallback to context.object
            o = context.object
        if self.archipack_filter(o):
            self.act = o
        return self.execute(context)

    def add_to_reference(self, context, o):
        # no reference when rendering in background
        if bpy.app.background:
            return

        ref = self.get_reference_point(self.act)
        if ref is None:
            ref = o
        with ensure_select_and_restore(context, ref, [o]):
            bpy.ops.archipack.add_reference_point()


class ArchipackDrawTool(ArchipackObjectsManager):
    """
        Draw tools
    """
    bl_category = 'Archipack'
    bl_options = {'INTERNAL', 'UNDO'}

    disabled_walls = {}
    optimize_walls = True

    def region_2d_to_orig_and_vect(self, context, event):

        region = context.region
        rv3d = context.region_data
        coord = (event.mouse_region_x, event.mouse_region_y)
        vec = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
        orig = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)

        return rv3d.is_perspective, orig, vec

    def isect_vec_plane(self, p0, vec, p_co, p_no, epsilon=1e-6):
        """
        p0, vec: define the line
        p_co, p_no: define the plane:
            p_co is a point on the plane (plane coordinate).
            p_no is a normal vector defining the plane direction;
                 (does not need to be normalized).
        return a Vector or None (when the intersection can't be found).
        """
        u = vec
        d = p_no.dot(u)

        if abs(d) > epsilon:
            w = p0 - p_co
            t = -p_no.dot(w) / d
            return p0 + t * u
        else:
            # The segment is parallel to plane
            return None

    def mouse_to_plane(self, context, event, origin=Vector((0, 0, 0)), normal=Vector((0, 0, 1))):
        """
            convert mouse pos to 3d point over plane defined by origin and normal
            return None if the point is behind camera view
        """
        is_perspective, orig, vec = self.region_2d_to_orig_and_vect(context, event)
        # pt = intersect_line_plane(orig, orig + vec, origin, normal, False)
        pt = self.isect_vec_plane(orig, vec, origin, normal)

        # fix issue with parallel plane (front or left ortho view)
        if pt is None:
            pt = self.isect_vec_plane(orig, vec, origin, Vector((0, 1, 0)))

        if pt is None:
            pt = self.isect_vec_plane(orig, vec, origin, Vector((1, 0, 0)))
            # pt = intersect_line_plane(orig, orig + vec, origin, vec, False)

        if pt is None:
            return None

        if is_perspective:
            # Check if point is behind point of view (mouse over horizon)
            y = Vector((0, 0, 1))
            # x = vec.cross(y)
            x = y.cross(vec)
            itM = Matrix([
                [x.x, y.x, vec.x, orig.x],
                [x.y, y.y, vec.y, orig.y],
                [x.z, y.z, vec.z, orig.z],
                [0, 0, 0, 1]
                ]).inverted()
            res = itM @ pt

            if res.z < 0:
                return None

        return pt

    def is_snapping(self, context, event, pt):
        """
        :param context:
        :param event:
        :param pt: point in world space
        :return:
        """
        is_perspective, orig, vec = self.region_2d_to_orig_and_vect(context, event)
        if vec.x != 0 and vec.y != 0:
            x = vec.cross(Vector((0, 0, 1)))
        else:
            x = vec.cross(Vector((1, 0, 0)))
        y = x.cross(vec)
        sp = Matrix([
            [x.x, y.x, vec.x, orig.x],
            [x.y, y.y, vec.y, orig.y],
            [x.z, y.z, vec.z, orig.z],
            [0, 0, 0, 1]
        ]).inverted() @ pt
        return abs(sp.x) > 0.001 or abs(sp.y) > 0.001

    def mouse_to_scene_raycast(self, context, event):
        """
            convert mouse pos to 3d point over plane defined by origin and normal
        """
        is_perspective, orig, vec = self.region_2d_to_orig_and_vect(context, event)
        res, pos, normal, face_index, obj, matrix_world = self.scene_ray_cast(
            context,
            orig,
            vec)

        return res, pos, normal, face_index, obj, matrix_world
    
    def restore_walls(self, context):
        """
         Enable finish on wall
        """
        last = context.active_object
        for name in self.disabled_walls.keys():
            # print("restore_walls ", name)
            o = context.scene.objects.get(name.strip())
            if o is not None:
                d = o.data.archipack_wall2[0]
                self.select_object(context, o, True)
                d.finish_enable = True
                self.unselect_object(context, o)
        self.select_object(context, last, True)
        self.disabled_walls.clear()

    def mouse_hover_wall(self, context, event):
        """
            convert mouse pos to matrix at bottom of surrounded wall, y oriented outside wall
        """
        res, pt, y, i, o, tM = self.mouse_to_scene_raycast(context, event)

        if res and o.data is not None:

            z = Vector((0, 0, 1))
            y = -y
            x = y.cross(z)
            width = 0
            z_offset = 0
            res = False
            if 'archipack_wall2' in o.data :

                res = True

                d = o.data.archipack_wall2[0]
                width = d.width
                z_offset = d.z_offset

                # TODO: identify wall segment and pass finishing overflow_out
                
                if self.optimize_walls and d.finish_enable and (len(d.finish) > 0 or len(d.finish2) > 0):
                    # print("Disable finish for %s" % o.name)
                    with ensure_select_and_restore(context, o, [o]):
                        d.finish_enable = False
                    self.disabled_walls[o.name] = o

            elif ('archipack_wall' in o.data or 
                'archipack_custom_wall' in o):

                # one point on the opposite to raycast side (1 unit (meter) inside)
                p0 = pt + y.normalized()
                # direction
                dp = -y.normalized()
                # cast another ray to find wall depth
                attempt = 0
                while res is False and attempt < 5:
                    res, pos, normal, face_index, obj, matrix_world = self.scene_ray_cast(context, p0, dp)
                    # must hit another face or another object
                    res = res and (i != face_index or obj.name != o.name)
                    p0 -= dp
                    attempt += 1

                if res:
                    width = (pt - pos).to_2d().length

            if res:
                pt += (0.5 * width) * y.normalized()
                return True, Matrix([
                    [x.x, y.x, z.x, pt.x],
                    [x.y, y.y, z.y, pt.y],
                    [x.z, y.z, z.z, o.matrix_world.translation.z],
                    [0, 0, 0, 1]
                ]), o, width, y, z_offset

        return False, Matrix(), None, 0, Vector(), 0


    """
    
    
    def copy_material(self, context, o, d, _int, _ext):
        if _ext is not None:
            if _ext.material_override:
                id_mat = _ext.id_mat(MAT_PART_OUTSIDE)
            else:
                id_mat = d.id_mat(MAT_PART_OUTSIDE)
            self.mat_name = o.data.materials[id_mat].name
        if _int is not None:
            if _int.material_override:
                id_mat = _int.id_mat(MAT_PART_INSIDE)
            else:
                id_mat = d.id_mat(MAT_PART_INSIDE)
            self.mat_name = o.data.materials[id_mat].name
        
    def paste_material(self, context, o, d, _int, _ext):
        
        # Find if material is on the wall
        id_mat = None
        for i, mat in o.data.materials:
            if mat is not None and mat.name == self.mat_name:
                id_mat = i
                break
                
        if id_mat is None:
            id_mat = len(o.data.materials)
            o.data.materials.append(bpy.data.materials[self.mat_name])
            
        if _ext is not None:
            if not _int.override_material:
                _int.idmat[MAT_PART_INSIDE] = d.idmat[MAT_INSIDE]
            _ext.idmat[MAT_PART_OUTSIDE] = id_mat
            _ext.override_material = True
          
        if _int is not None:
            if not _ext.override_material:
                _ext.idmat[MAT_PART_OUTSIDE] = d.idmat[MAT_OUTSIDE]
            _int.idmat[MAT_PART_INSIDE] = id_mat
            _int.override_material = True
     
    def wall_material(self, context, event):
    
        o, d, _int, _ext, _seg, _t = self.wall_segment_under_mouse(self, context, event)
        if o is not None:
            if event.alt:
                self.copy_material(context, o, d, _int, _ext)
            else:
                self.paste_material(context, o, d, _int, _ext)
           
    def wall_segment_under_mouse(self, context, event):
    
        # Paste a material index over wall
        res, pt, y, i, o, tM = self.mouse_to_scene_raycast(context, event)

        if res and o.data is not None:

            # When pasting over other object, use face material index to identify slot

            if 'archipack_wall2' in o.data:

                d = o.data.archipack_wall2[0]
                g = d.get_generator(o)
                last = 1e32
                _seg = None
                id_mat = -1
                _ext = None
                _int = None
                _t = -1
                for _k, part in zip(g.outside.valid_segs, d.parts):
                    res, dist, t = _k.point_sur_segment(pt)
                    if res and abs(dist) < last:
                        last = abs(dist)
                        _seg = _k
                        _ext = part
                        _t = t
                  
                if last > 0.001:
                    for _k, part in zip(g.inside.valid_segs, d.parts):
                        res, dist, t = _k.point_sur_segment(pt)
                        if res and abs(dist) < last:
                            last = abs(dist)
                            _seg = _k
                            _int = part
                            _t = t
                            
                return o, d, _int, _ext, _seg, _t
                
            elif ('archipack_wall' in o.data or
                  'archipack_custom_wall' in o):
        
        return None, None, None, None, None, None
    """


class ArchipackArrayTool(ArchipackCreateTool):

    x: FloatProperty(
        name='Spacing',
        unit='LENGTH', subtype='DISTANCE',
        default=2.0, precision=5,
        description='Spacing, positive on the left'
    )
    count: IntProperty(
        name="Duplicates",
        min=2,
        default=2
    )
    first_location: FloatProperty(
        name='Offset',
        unit='LENGTH', subtype='DISTANCE',
        default=0.0, precision=5,
        description='Current offset from segment start'
    )
    objects_size: FloatProperty(
        name='Overall width',
        unit='LENGTH', subtype='DISTANCE',
        default=0.0, precision=5,
        description='Overall width of objects on segment'
    )
    available_space: FloatProperty(
        name='Avaliable',
        unit='LENGTH', subtype='DISTANCE',
        default=0.0, precision=5,
        description='Spacing, positive on the left'
    )
    gap_start: FloatProperty(
        name='Gap start',
        unit='LENGTH', subtype='DISTANCE',
        default=0.0, precision=5,
        description='Spacing, positive on the left'
    )
    gap_end: FloatProperty(
        name='Gap end',
        unit='LENGTH', subtype='DISTANCE',
        default=0.0, precision=5,
        description='Spacing, positive on the left'
    )
    limit_count: BoolProperty(
        name="Limit count",
        default=True
    )
    mode: EnumProperty(
        name="Mode",
        items=(
            ("START_END_COUNT", "(Auto) Start - End - Count", "Auto equal spacing"),
            ("COUNT", "(Auto) Count", "Equal space, half space on borders"),
            ("SPACING", "(Border) Spacing", "Auto equal space on borders"),
            ("START_COUNT_SPACING", "(Border) Start - Spacing - Count", ""),
            ("END_COUNT_SPACING", "(Border) End - Spacing - Count", ""),
            ("CUR_COUNT_SPACING", "(Current, Border) Spacing - Count", "About current location"),
            ("AXIS_SPACING", "(Axis) Spacing", "Auto equal space on borders"),
            ("AXIS_START_COUNT_SPACING", "(Axis) Start - Spacing- Count", ""),
            ("AXIS_END_COUNT_SPACING", "(Axis) End - Spacing - Count ", ""),
            ("AXIS_CUR_COUNT_SPACING", "(Current, Axis) Spacing - Count", "About current location")
        )
    )

    def filter_object(self, o):
        raise NotImplementedError

    def get_object_datablock(self, o):
        raise NotImplementedError

    def make_array(self, context, sel):
        startplace, spacing, count = self.distribute_objects()
        if count < 2:
            return
        link = len(self.segment_t) < 2
        for o in sel:
            d = self.get_object_datablock(o)
            if d is not None:
                # find walls this hole belongs
                hole = d.find_hole(o)
                ref = self.get_reference_point(o)
                # get custom wall, wall and wall2
                objs = self.filter_selection_loose(ref.children)
                walls_holes = set()
                for key, walls in objs.items():
                    if "wall" in key:
                        for name in walls:
                            c = self.get_scene_object(context, name)
                            m = c.modifiers.get("AutoMixedBoolean")
                            if m is not None and m.object is not None:
                                basis = m.object
                                for h_mod in basis.modifiers:
                                    if h_mod.type == "BOOLEAN" and h_mod.object == hole:
                                        walls_holes.add(basis)

                for i in range(count - 1):
                    c = self.duplicate_object(context, o, link)
                    # p = o.matrix_world @ Vector((startplace + (i + 1) * spacing, 0, 0))
                    tM = self._interp(startplace + (i + 1) * spacing)
                    c.parent = ref
                    c.matrix_world = tM
                    # ref.matrix_world @
                    # c.location = ref.matrix_world.inverted() @ p
                    # c.matrix_world.translation = p

                    hole = d.find_hole(c)
                    if hole is not None:
                        # unlink hole
                        hole.data = hole.data.copy()
                        for basis in walls_holes:
                            m = basis.modifiers.new('AutoMerge', 'BOOLEAN')
                            m.operation = 'UNION'
                            m.object = hole
                        hole.hide_set(state=True)

                tM = self._interp(startplace)
                o.matrix_world = tM
                # o.matrix_local = ref.matrix_world.inverted() @ tM
                # p = o.matrix_world @ Vector((startplace, 0, 0))ref.matrix_world @
                # o.location = ref.matrix_world.inverted() @ p
                # o.matrix_world.translation = p

    def draw(self, context):
        layout = self.layout
        layout.prop(self, "mode", text_ctxt="archipack")

        if "START" in self.mode:
            layout.prop(self, "gap_start", text_ctxt="archipack")

        if "END" in self.mode:
            layout.prop(self, "gap_end", text_ctxt="archipack")

        if "COUNT" in self.mode:
            layout.prop(self, "count", text_ctxt="archipack")
            layout.prop(self, "limit_count", text_ctxt="archipack")

        if "SPACING" in self.mode:
            layout.prop(self, "x", text_ctxt="archipack")

    # -----------------------------------------------------
    # Execute
    # -----------------------------------------------------
    def execute(self, context):
        if context.mode == "OBJECT":
            # select and make active
            bpy.ops.archipack.disable_manipulate()
            sel = context.selected_objects[:]
            o = context.active_object
            with ensure_select_and_restore(context, o, sel):
                self.make_array(context, sel)
            wall = self.get_scene_object(context, self.wall_name)
            with ensure_select_and_restore(context, wall, [wall]):
                wd = wall.data.archipack_wall2[0]
                wd.update(context, setup_childs=True, relocate_childs=True, allow_throttle=False)

            self.select_object(context, o, True)

            return {'FINISHED'}
        else:
            self.report({'WARNING'}, "Archipack: Option only valid in Object mode")
            return {'CANCELLED'}

    def distribute_objects(self):

        startplace, spacing, count = 0, self.x, self.count

        if self.mode == "START_END_COUNT":
            startplace = self.gap_start + self.first_location
            available = self.available_space - (self.gap_end + self.gap_start)
            max_count = floor(available / self.objects_size)
            if self.limit_count:
                count = min(count, max_count)
            emptyspace = available - (count * self.objects_size)
            gap = emptyspace / (count - 1)
            spacing = self.objects_size + gap

        elif self.mode == "AXIS_SPACING":
            n_spaces = floor((self.available_space - self.objects_size) / spacing)
            startplace = self.first_location + 0.5 * (self.available_space - n_spaces * spacing - self.objects_size)
            count = n_spaces + 1

        elif self.mode == "SPACING":
            spacing = self.objects_size + self.x
            n_spaces = floor((self.available_space - self.objects_size) / spacing)
            count = n_spaces + 1
            gap = self.available_space - (n_spaces * spacing + self.objects_size)
            startplace = self.first_location + 0.5 * gap

        elif self.mode == "COUNT":
            max_count = floor(self.available_space / self.objects_size)
            if self.limit_count:
                count = min(count, max_count)
            emptyspace = self.available_space - (count * self.objects_size)
            if count > 0:
                gap = emptyspace / count
                spacing = self.objects_size + gap
                startplace = self.first_location + 0.5 * gap

        elif self.mode == "AXIS_START_COUNT_SPACING":
            startplace = self.gap_start + self.first_location - 0.5 * self.objects_size
            n_spaces = floor((self.available_space - self.objects_size - self.gap_start) / spacing)
            if self.limit_count:
                count = min(count, n_spaces + 1)

        elif self.mode == "START_COUNT_SPACING":
            spacing = self.objects_size + self.x
            startplace = self.gap_start + self.first_location
            available = self.available_space - self.gap_start
            max_count = floor(available / spacing)
            if self.limit_count:
                count = min(count, max_count)

        elif self.mode == "AXIS_END_COUNT_SPACING":
            n_spaces = floor((self.available_space - self.objects_size - self.gap_end) / spacing)
            if self.limit_count:
                count = min(count, n_spaces + 1)
            startplace = self.available_space + self.first_location - count * spacing - 0.5 * self.objects_size

        elif self.mode == "END_COUNT_SPACING":
            spacing = self.objects_size + self.x
            available = self.available_space - self.gap_end
            max_count = floor(available / spacing)
            if self.limit_count:
                count = min(count, max_count)
            startplace = self.available_space + self.first_location - count * spacing

        elif self.mode == "AXIS_CUR_COUNT_SPACING":
            n_spaces = floor((self.available_space + self.first_location - self.objects_size) / spacing)
            if self.limit_count:
                count = min(count, n_spaces + 1)

        elif self.mode == "CUR_COUNT_SPACING":
            spacing = self.objects_size + self.x
            available = self.available_space
            max_count = floor(available / spacing)
            if self.limit_count:
                count = min(count, max_count)

        print(startplace, spacing, count)

        return startplace, spacing, count

    def _interp(self, x):
        x0, _s = self.segment_t[-1]
        j = len(self.segment_t) - 1
        while x < x0 and j > 0:
            j -= 1
            x0, _s = self.segment_t[j]

        sx = x - x0
        dx = _s.length
        if dx > 0:
            t = sx / dx
        else:
            t = 0
        p = _s.lerp(t)
        vx = -_s.v_normalized
        vy = vx.cross(Z_AXIS)
        p += vy * self.y
        return Matrix([
            [vx.x, vy.x, 0, p.x],
            [vx.y, vy.y, 0, p.y],
            [vx.z, vy.z, 1, 0],
            [0, 0, 0, 1]
        ])

    def find_wall_child(self, wd, o):
        for c in wd.childs:
            if c.child_name == o.name:
                return c
        return None

    def find_wall(self, context, o):
        ref = self.get_reference_point(o)
        if ref is not None:
            objs = self.filter_selection_loose(ref.children)
            if "archipack_wall2" in objs:
                for name in objs['archipack_wall2']:
                    wall = self.get_scene_object(context, name)
                    wd = wall.data.archipack_wall2[0]
                    wg = wd.get_generator(wall)
                    child = self.find_wall_child(wd, o)
                    if child is not None:
                        return wall, wd, wg, child

        return None, None, None, -1

    def invoke(self, context, event):
        o = context.active_object
        sel = [c for c in context.selected_objects if self.filter_object(c)]
        d = self.get_object_datablock(o)
        # find wall segment for the active object
        wall, wd, wg, child = self.find_wall(context, o)
        if wall is not None:
            self.wall_name = wall.name
            self.y = 0.5 * wd.width
            _seg_idx = child.wall_idx

            # handle multiple collinear segments - curved segments < ~15 deg
            _seg = wg.outside.segs[_seg_idx]
            # find first seg on left side
            while _seg.has_last and _seg_idx > 0 and abs(_seg.delta_angle(_seg._last)) < 0.2:
                _seg_idx -= 1
                _seg = _seg._last
            first_segment = _seg_idx

            # print("first_segment", first_segment)

            # find last seg on right part
            _segs_idx = set()
            _segs_idx.add(_seg_idx)
            available_space = _seg.length
            # absolute distance from wall start
            segment_len = 0
            segment_param_t = {}
            segment_param_t[0] = 0
            for i, _s in enumerate(wg.outside.valid_segs):
                segment_len += _s.length
                segment_param_t[i + 1] = segment_len
            # store segment dist map
            segment_t = []
            segment_t.append((segment_param_t[_seg_idx], _seg))
            n_segs = len(wg.outside.valid_segs) - 1
            while _seg.has_next and _seg_idx < n_segs and abs(
                    _seg._next.delta_angle(_seg)) < 0.2:
                available_space += _seg.length
                _seg_idx += 1
                # segment_param_t[_seg_idx] = available_space
                _segs_idx.add(_seg_idx)
                _seg = _seg._next
                segment_t.append((segment_param_t[_seg_idx], _seg))

            self.segment_t = segment_t

            logger.debug("available_space %s" % available_space)

            logger.debug("segment_param_t %s" % segment_param_t)

            logger.debug("segment_t %s" % segment_t)

            childs = []
            childs.append((o, segment_param_t[child.wall_idx] + child.pos.x, d.x))
            for c in sel:
                # ensure other objects are on same segment
                if c.name != o.name:
                    child = self.find_wall_child(wd, c)
                    if child.wall_idx in _segs_idx:
                        d = self.get_object_datablock(c)
                        t = segment_param_t[child.wall_idx] + child.pos.x
                        childs.append((c, t, d.x))
            # sort by distance from start
            childs.sort(key=lambda x: x[1])

            # offset so the first border is at wall segment start
            # self.first_location = (0.5 * childs[0][2]) - childs[0][1]
            self.first_location = segment_param_t[first_segment] + (0.5 * childs[0][2])
            # max distance between childs border - border
            self.objects_size = childs[0][1] - childs[-1][1] + 0.5 * (childs[0][2] + childs[-1][2])
            # self.available_space = _seg.length
            self.available_space = available_space
            return context.window_manager.invoke_props_dialog(self)
        else:
            self.report({'WARNING'}, "Must draw object over wall before using array")
        return {'CANCELLED'}

