# -*- coding:utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

# ----------------------------------------------------------
# Author: Stephen Leger (s-leger)
#
# ----------------------------------------------------------
import bpy
from bpy.types import Operator
from math import sin, cos, pow
from .bmesh_utils import BmeshEdit as bmed
from .archipack_abstraction import ensure_select_and_restore
from .archipack_object import ArchipackObjectsManager
import logging

logger = logging.getLogger(__package__)


class ARCHIPACK_OT_convert_all_to_tris(ArchipackObjectsManager, Operator):
    bl_idname = "archipack.convert_all_to_tris"
    bl_label = "Triangulate"
    bl_description = "Triangulate mesh for visible objects"
    bl_options = {"REGISTER"}

    def execute(self, context):
        objects = self.get_scene_objects(context)
        for o in objects:
            if o.type == 'MESH' and o.visible_get():
                with ensure_select_and_restore(context, o, [o], object_mode='EDIT'):
                    bpy.ops.mesh.select_all(action="SELECT")
                    bpy.ops.mesh.quads_convert_to_tris(quad_method='BEAUTY', ngon_method='BEAUTY')

        return {'FINISHED'}


class ARCHIPACK_OT_remove_vertex_colors(ArchipackObjectsManager, Operator):
    bl_idname = "archipack.remove_vertex_colors"
    bl_label = "Remove vertex colors"
    bl_description = "Remove vertex colors for visible objects"
    bl_options = {"REGISTER"}

    def execute(self, context):
        objects = self.get_scene_objects(context)
        for o in objects:
            if o.type == 'MESH' and o.visible_get():
                vg = o.data.vertex_colors.get("Archipack")
                if vg:
                    o.data.vertex_colors.remove(vg)
        return {'FINISHED'}


class ARCHIPACK_OT_apply_random_uv(ArchipackObjectsManager, Operator):
    bl_idname = "archipack.apply_random_uv"
    bl_label = "Apply random uv"
    bl_description = "Apply random vertex colors to mesh uvs for visible objects"
    bl_options = {"REGISTER", "UNDO"}

    @staticmethod
    def rotate_uv(a, u, v):
        ca = cos(a)
        sa = sin(a)
        return (ca * u) + (sa * v), (ca * v) - (sa * u)

    @staticmethod
    def translate_uv(u, v, du, dv):
        return u + du, v + dv

    @staticmethod
    def scale_uv(u, v, su, sv):
        return u * su, v * sv

    @staticmethod
    def has_random_uvs(o):
        for matid, slot in enumerate(o.material_slots):
            mat = slot.material
            if mat is not None and mat.use_nodes:
                nodes = mat.node_tree.nodes
                for node in nodes:
                    if node.type == "GROUP" and "Archipack Random UV" in node.node_tree.name:
                        return True
        return False

    def gamma_correct(self, gamma, r, g, b):
        rc = pow(r, gamma)
        gc = pow(g, gamma)
        bc = pow(b, gamma)
        return rc, gc, bc

    def apply(self, o):
        if self.has_random_uvs(o):
            bm = bmed._start(o)
            for matid, slot in enumerate(o.material_slots):
                mat = slot.material
                if mat is not None and mat.use_nodes:
                    tree = mat.node_tree
                    nodes = tree.nodes
                    n = None
                    # flag vertex colors output not linear due to internal gamma curve
                    is_not_linear = True
                    for node in nodes:
                        if node.type == "GROUP" and "Archipack Random UV" in node.node_tree.name:
                            n = node
                            for sub in n.node_tree.nodes:
                                if sub.type == 'GAMMA':
                                    is_not_linear = False
                                    break
                            break
                    if n is not None:
                        scale_u = n.inputs[0].default_value
                        scale_v = n.inputs[1].default_value
                        # use same factor than in node (approx 2 * pi)
                        rotation_uv = n.inputs[2].default_value / 360.0 * 6.28
                        random_rotation = n.inputs[3].default_value / 100.0 * 6.28
                        # disconnect uv output
                        for link in n.outputs[3].links:
                            tree.links.remove(link)
                        n.inputs[0].default_value = scale_u
                        vlay = bm.loops.layers.color.get("Archipack")
                        # Mimic Archipack Random UV group effect
                        if vlay is not None:
                            layer = bm.loops.layers.uv.verify()
                            for face in bm.faces:
                                if face.material_index == matid:
                                    for loop in face.loops:
                                        u, v = loop[layer].uv
                                        r, g, b, a = loop[vlay]
                                        # Vertex colors are gamma corrected in shader to get linear result
                                        if is_not_linear:
                                            r, g, b = self.gamma_correct(2.2, r, g, b)
                                        u, v = self.rotate_uv(rotation_uv, u, v)
                                        u, v = self.translate_uv(u, v, g / scale_u, b / scale_v)
                                        u, v = u * scale_u, v * scale_v
                                        u, v = self.rotate_uv(random_rotation * r, u, v)
                                        # print("rotate_uv", u, v, random_rotation * r)
                                        loop[layer].uv = (u, v)
                                        # keep vertex colors
                                        # loop[vlay] = (r, 0, 0, 1)
            bmed._end(bm, o)

    def execute(self, context):

        objects = self.get_scene_objects(context)
        for o in objects:
            if o.type == 'MESH' and o.visible_get():
                self.apply(o)

        return {'FINISHED'}


def register():
    bpy.utils.register_class(ARCHIPACK_OT_apply_random_uv)
    bpy.utils.register_class(ARCHIPACK_OT_convert_all_to_tris)
    bpy.utils.register_class(ARCHIPACK_OT_remove_vertex_colors)


def unregister():
    bpy.utils.unregister_class(ARCHIPACK_OT_apply_random_uv)
    bpy.utils.unregister_class(ARCHIPACK_OT_convert_all_to_tris)
    bpy.utils.unregister_class(ARCHIPACK_OT_remove_vertex_colors)